/*
 *  contdata.c  -  manipulate the contour data 
 */

#include <stdio.h>
#include <strings.h>
#include <math.h>
#include "plot2D.h"

/* 
 * The contour data is originally in a 1D array
 *   [ [0,0]...[i,0]...[nx,0], [0,j]...[i,j]...[nx,ny] ]
 * Go thru the contour data and rearrange it into a set of triangles
 *
 * This data is stored in a linked list, since there is there could be
 * multiple contour data sets (e.g. 2 sets of data from 2 files)
 */
CNplotsetptr 
CNget_rectilinear_contour_data(filename,
                               xmin,xmax,ymin,ymax,zmin,zmax,
                               data_array,nx,ny,
                               joincurve,
                               cstep,nsteps,linetypes,zlog,ID)
char      *filename;          /* Name of original file */
double    xmin,xmax;          /* Data boundaries (x)   */
double    ymin,ymax;          /* Data boundaries (y)   */
double    zmin,zmax;          /* Data boundaries (z)   */
double    *data_array;        /* 1D data array f(x,y)  */
int       nx, ny;             /* #x and y gridpoints   */
int       joincurve;          /* Contour interpolation */
double    cstep;              /* Contour step size     */
int       nsteps;             /* No of contour steps   */
int       linetypes;          /* No. contour linetypes */
int       zlog;               /* Logarithmic z-values  */
int       ID;                 /* Plot ID               */
{
   static void   convert_rect_to_tria();
   CNplotsetptr  CNmake_plotset();
   CNplotsetptr  pptr;

   /* 
    * Create and initialize a data structure to hold this contour data set
    */
   if ((pptr = CNmake_plotset(filename,CN_CONTOUR,
                              xmin,xmax,ymin,ymax,zmin,zmax,
                              cstep,nsteps,linetypes,zlog,ID))==NULL)
      return(pptr);

   /*
    * convert the rectangular array into a linked list of triangles 
    */
   convert_rect_to_tria(pptr,data_array,nx,ny,joincurve);

   /* 
    * Return the pointer 
    */
   return(pptr);
}


/* 
 * The contour data is originally in a 3D list of triangles
 *   [ T0 (npts, pt0, pt1, pt2) ... TN (npts, pt0, pt1, pt2) ]
 * Go thru the contour data and rearrange it into a set of triangles
 *
 * This data is stored in a linked list, since there is there could be
 * multiple contour data sets (e.g. 2 sets of data from 2 files)
 *
 * xmin, xmax, ymin, ymax, zmin, zmax - bounds
 * triahead,triatail - pointers to linked list of triangles 
 * cstep  - the contour step size
 */
CNplotsetptr 
CNget_triangular_contour_data(filename,
                              xmin,xmax,ymin,ymax,zmin,zmax,
                              triahead,triatail,
                              cstep,nsteps,linetypes,zlog,ID)
char      *filename;          /* Name of original file */
double    xmin,xmax;          /* Data boundaries (x)   */
double    ymin,ymax;          /* Data boundaries (y)   */
double    zmin,zmax;          /* Data boundaries (z)   */
CNtriaptr triahead,triatail;  /* List of triangles     */
double    cstep;              /* Contour step size     */
int       nsteps;             /* No of contour steps   */
int       linetypes;          /* No. contour linetypes */
int       zlog;               /* Logarithmic z-values  */
int       ID;                 /* Plot ID               */
{
   CNplotsetptr  CNmake_plotset();
   CNplotsetptr  pptr;

   /* 
    * Create and initialize a data structure to hold this contour data set
    */
   if ((pptr = CNmake_plotset(filename,CN_CONTOUR,
                              xmin,xmax,ymin,ymax,zmin,zmax,
                              cstep,nsteps,linetypes,zlog,ID))==NULL)
      return(pptr);

   /*
    * Link the triangles to the data-structure 
    */
   pptr->triahead = triahead;
   pptr->triatail = triatail;

   /* 
    * Return the pointer 
    */
   return(pptr);
}


/*
 * convert the rectangular array into a linked list of triangles
 */
static void convert_rect_to_tria(Pptr,data_array,nx,ny,joincurve)
CNplotsetptr Pptr;
double       *data_array;
int          nx, ny;
int          joincurve;
{
   static void conv_rect_struct();
   CNtriaptr   CNinsert_tria();
   double      CNget_1D_double_array_value();

   CNdatapt    pt00, pt01, pt10, pt11;
   double      xmin, xmax, ymin, ymax, zmin, zmax;
   double      dx, dy, delta, ztarget;
   int         i, j;

   /* Initialize the boundaries */
   xmin = Pptr->xmin;
   xmax = Pptr->xmax;
   ymin = Pptr->ymin;
   ymax = Pptr->ymax;
   zmin = Pptr->zmin;
   zmax = Pptr->zmax;
   dx   = (xmax-xmin)/(double)(nx-1);
   dy   = (ymax-ymin)/(double)(ny-1);

   /* Go thru the rectangular array and break up into triangles */
   for (i=0; i<nx-1; i++)
   for (j=0; j<ny-1; j++) {
      /* get the xyz coordinates for the four points at the corners */
      pt00.x = xmin + (i  )*dx;
      pt00.y = ymin + (j  )*dy;
      pt00.z = CNget_1D_double_array_value(data_array,((i  )+(j  )*nx),nx*ny);
      pt10.x = xmin + (i+1)*dx;
      pt10.y = ymin + (j  )*dy;
      pt10.z = CNget_1D_double_array_value(data_array,((i+1)+(j  )*nx),nx*ny);
      pt01.x = xmin + (i  )*dx;
      pt01.y = ymin + (j+1)*dy;
      pt01.z = CNget_1D_double_array_value(data_array,((i  )+(j+1)*nx),nx*ny);
      pt11.x = xmin + (i+1)*dx;
      pt11.y = ymin + (j+1)*dy;
      pt11.z = CNget_1D_double_array_value(data_array,((i+1)+(j+1)*nx),nx*ny);

      /* save the structure */
      conv_rect_struct(Pptr,&(pt00),&(pt01),&(pt10),&(pt11));
   }

   /* 
    * If the contours are to be joined at the boundary, then
    * there must be a boundary layer around the contour box.
    * Create a bunch of triangles filling up this boundary layer.
    */
   if (joincurve != CN_NONE) {
      /* determine if the boundary layer is a min or a maxima */
      if (joincurve == CN_HIGH)
         ztarget = zmax;
      else
         ztarget = zmin;

      if ((xmax-xmin) > (ymax-ymin)) 
         delta = 1.0e-4 * (ymax-ymin);
      else   
         delta = 1.0e-4 * (xmax-xmin);

      /* fill a boundary layer for i=0, variable j */
      for (j=0; j<ny-1; j++) {
         i=0;
         pt00.x= xmin + (i  )*dx;
         pt00.y= ymin + (j  )*dy;
         pt00.z= CNget_1D_double_array_value(data_array,((i  )+(j  )*nx),nx*ny);
         pt01.x= xmin + (i  )*dx;
         pt01.y= ymin + (j+1)*dy;
         pt01.z= CNget_1D_double_array_value(data_array,((i  )+(j+1)*nx),nx*ny);
         pt10.x= pt00.x - delta; 
         pt10.y= pt00.y; 
         pt10.z= ztarget;
         pt11.x= pt01.x - delta; 
         pt11.y= pt01.y; 
         pt11.z= ztarget;

         /* save the structure */
         conv_rect_struct(Pptr,&(pt00),&(pt01),&(pt10),&(pt11));

         i=nx-1;
         pt00.x= xmin + (i  )*dx;
         pt00.y= ymin + (j  )*dy;
         pt00.z= CNget_1D_double_array_value(data_array,((i  )+(j  )*nx),nx*ny);
         pt01.x= xmin + (i  )*dx;
         pt01.y= ymin + (j+1)*dy;
         pt01.z= CNget_1D_double_array_value(data_array,((i  )+(j+1)*nx),nx*ny);
         pt10.x= pt00.x + delta; 
         pt10.y= pt00.y; 
         pt10.z= ztarget;
         pt11.x= pt01.x + delta; 
         pt11.y= pt01.y; 
         pt11.z= ztarget;

         /* save the structure */
         conv_rect_struct(Pptr,&(pt00),&(pt01),&(pt10),&(pt11));
      }

      /* fill a boundary layer for j=0, variable i */
      for (i=0; i<nx-1; i++) {
         j=0;
         pt00.x= xmin + (i  )*dx;
         pt00.y= ymin + (j  )*dy;
         pt00.z= CNget_1D_double_array_value(data_array,((i  )+(j  )*nx),nx*ny);
         pt10.x= xmin + (i+1)*dx;
         pt10.y= ymin + (j  )*dy;
         pt10.z= CNget_1D_double_array_value(data_array,((i+1)+(j  )*nx),nx*ny);
         pt01.x= pt00.x; 
         pt01.y= pt00.y - delta; 
         pt01.z= ztarget;
         pt11.x= pt10.x; 
         pt11.y= pt10.y - delta; 
         pt11.z= ztarget;

         /* save the structure */
         conv_rect_struct(Pptr,&(pt00),&(pt01),&(pt10),&(pt11));

         j=ny-1;
         pt00.x= xmin + (i  )*dx;
         pt00.y= ymin + (j  )*dy;
         pt00.z= CNget_1D_double_array_value(data_array,((i  )+(j  )*nx),nx*ny);
         pt10.x= xmin + (i+1)*dx;
         pt10.y= ymin + (j  )*dy;
         pt10.z= CNget_1D_double_array_value(data_array,((i+1)+(j  )*nx),nx*ny);
         pt01.x= pt00.x; 
         pt01.y= pt00.y + delta; 
         pt01.z= ztarget;
         pt11.x= pt10.x; 
         pt11.y= pt10.y + delta; 
         pt11.z= ztarget;

         /* save the structure */
         conv_rect_struct(Pptr,&(pt00),&(pt01),&(pt10),&(pt11));
      }

      /* Now do the four triangles at the corners */
      i=0; j=0;
      pt00.x = xmin + (i  )*dx;
      pt00.y = ymin + (j  )*dy;
      pt00.z = CNget_1D_double_array_value(data_array,((i  )+(j  )*nx),nx*ny);
      pt01.x = pt00.x - delta;
      pt01.y = pt00.y ;
      pt01.z = ztarget;
      pt10.x = pt00.x;
      pt10.y = pt00.y - delta;
      pt10.z = ztarget;
      CNinsert_tria(&(Pptr->triahead),&(Pptr->triatail),&(pt00),&(pt10),&(pt01));
      i=nx-1; j=0;
      pt00.x = xmin + (i  )*dx;
      pt00.y = ymin + (j  )*dy;
      pt00.z = CNget_1D_double_array_value(data_array,((i  )+(j  )*nx),nx*ny);
      pt01.x = pt00.x + delta;
      pt01.y = pt00.y ;
      pt01.z = ztarget;
      pt10.x = pt00.x;
      pt10.y = pt00.y - delta;
      pt10.z = ztarget;
      CNinsert_tria(&(Pptr->triahead),&(Pptr->triatail),&(pt00),&(pt10),&(pt01));
      i=0; j=ny-1;
      pt00.x = xmin + (i  )*dx;
      pt00.y = ymin + (j  )*dy;
      pt00.z = CNget_1D_double_array_value(data_array,((i  )+(j  )*nx),nx*ny);
      pt01.x = pt00.x - delta;
      pt01.y = pt00.y ;
      pt01.z = ztarget;
      pt10.x = pt00.x;
      pt10.y = pt00.y - delta;
      pt10.z = ztarget;
      CNinsert_tria(&(Pptr->triahead),&(Pptr->triatail),&(pt00),&(pt10),&(pt01));
      i=nx-1; j=ny-1;
      pt00.x = xmin + (i  )*dx;
      pt00.y = ymin + (j  )*dy;
      pt00.z = CNget_1D_double_array_value(data_array,((i  )+(j  )*nx),nx*ny);
      pt01.x = pt00.x + delta;
      pt01.y = pt00.y ;
      pt01.z = ztarget;
      pt10.x = pt00.x;
      pt10.y = pt00.y - delta;
      pt10.z = ztarget;
      CNinsert_tria(&(Pptr->triahead),&(Pptr->triatail),&(pt00),&(pt10),&(pt01));
   }
}

/* take four 3D points and create 4 triangles */
static void conv_rect_struct(Pptr,pt00,pt01,pt10,pt11)
CNplotsetptr Pptr;
CNdatapt *pt00, *pt01, *pt10, *pt11;
{
   CNtriaptr  CNinsert_tria();
   CNdatapt   midpt;

   /* get the xyz coordinate of the midpoint */
   midpt.x = 0.50*(pt00->x + pt10->x);
   midpt.y = 0.50*(pt00->y + pt01->y);
   midpt.z = 0.25*(pt00->z + pt01->z + pt10->z + pt11->z);

   /* create triangles */
   CNinsert_tria(&(Pptr->triahead),&(Pptr->triatail),pt00,pt10,&midpt);
   CNinsert_tria(&(Pptr->triahead),&(Pptr->triatail),pt10,pt11,&midpt);
   CNinsert_tria(&(Pptr->triahead),&(Pptr->triatail),pt11,pt01,&midpt);
   CNinsert_tria(&(Pptr->triahead),&(Pptr->triatail),pt01,pt00,&midpt);
}


/*
 * Go through the contour data sets and find the
 * largest boundary that will fit around all the contours
 */
void
CNset_contour_boundaries(plothead, plottail, xmin, xmax, ymin, ymax, zmin, zmax)
CNplotsetptr plothead, plottail;
double       *xmin, *xmax, *ymin, *ymax, *zmin, *zmax;
{
   CNplotsetptr C;

   /* check to see if there is any data */
   if (plothead == NULL) {
      fprintf(stderr,"   ***Error! No data in the linked list\n");
      exit(-1);
   }

   /* Initialize the boundary data */
   *xmin = plothead->xmin;
   *xmax = plothead->xmax;
   *ymin = plothead->ymin;
   *ymax = plothead->ymax;
   *zmin = plothead->zmin;
   *zmax = plothead->zmax;

   /* Go thru the plot sets and compare boundaries */
   for (C=plothead->next; C!=NULL; C=C->next) {
      if (C->xmin < *xmin) *xmin = C->xmin;
      if (C->xmax > *xmax) *xmax = C->xmax;
      if (C->ymin < *ymin) *ymin = C->ymin;
      if (C->ymax > *ymax) *ymax = C->ymax;
      if (C->zmin < *zmin) *zmin = C->zmin;
      if (C->zmax > *zmax) *zmax = C->zmax;
   }
}

