/*
 * contour.c - manipulate the contour data
 */

#include <stdio.h>
#include <strings.h>
#include <math.h>
#include "plot2D.h"


/****************************************************************/
/**********                                            **********/
/**********         General contour subroutines        **********/
/**********                                            **********/
/****************************************************************/

/* 
 * Guess the "best" value for the contour step size 
 * There should be about 10 contours - round cstep up to its
 * nearest decimal place (0.0156 -> 0.02, 0.014 -> 0.01)
 *
 * If cstep is specified (non-zero) then check it and use it
 * Otherwise, if nsteps is specified (non-zero) then check it and use it
 * otherwise, ask user for csteps
 */
void CNselect_contour_step(Pptr)
CNplotsetptr Pptr;
{
#define    CN_MAXSTEPS 50
#define    CN_IDLSTEPS 10
   double  CNround_to_decimal();
   double  cstep;
   double  zmin, zmax;
   int     nsteps = 0, recalc = CN_FALSE;
   char    c;

   /* check the contour step size */
   zmin   = Pptr->zmin;
   zmax   = Pptr->zmax;
   cstep  = Pptr->cstep;
   nsteps = Pptr->nsteps;

   /* Print out zmin and zmax as an aid */
   fprintf(stdout,"\n   Taking multiple slices of the dataset\n");
   fprintf(stdout,"      Dataset Information:\n");
   fprintf(stdout,"      PLOT ID   = %d\n",Pptr->ID);
   fprintf(stdout,"      Origin    = %s\n",Pptr->filename);
   fprintf(stdout,"      Linetypes = %d\n",Pptr->linetypes);
   fprintf(stdout,"      Z-Log     = %s\n",((Pptr->zlog) ? "TRUE" : "FALSE"));
   fprintf(stdout,"      Maximum z-value        = %f\n",zmax);
   fprintf(stdout,"      Minimum z-value        = %f\n",zmin);
   if (zmax < zmin) {
      fprintf(stderr,"   ***Error: zmax < zmin!\n");
      exit(-1);
   } 
   if (fabs(zmax - zmin) < CN_SMALL) {
      fprintf(stderr,"   ***Error: zmax = zmin!\n");
      fprintf(stderr,"      Cannot take contours of a horizontal plane!\n");
      exit(-1);
   } 

   /* if the contour step size has already been defined, check it */
   if (cstep > CN_SMALL) {
      nsteps = (zmax-zmin)/cstep;
      recalc = CN_FALSE;
      if (nsteps < 1) {
         fprintf(stdout,"   ***Error : Too few contours %d\n",nsteps);
         recalc = CN_TRUE;
      } else if (nsteps > CN_MAXSTEPS) {
         fprintf(stdout,"   ***Error : Too many contours %d\n",nsteps);
         recalc = CN_TRUE;
      } else {
         recalc = CN_FALSE;
      }
      /* Recalculate the contour step size */
      if (recalc) {
         fprintf(stdout,"   Recalculating the contour step size...\n");
         nsteps = CN_IDLSTEPS;
         cstep  = CNround_to_decimal( (zmax - zmin)/(double)nsteps );
         nsteps = (int) ((zmax-zmin)/cstep);
      }

   /* if the number of steps has already been defined, check it */
   } else if (nsteps > 0) {
      recalc = CN_FALSE;
      if (nsteps > CN_MAXSTEPS) {
         fprintf(stdout,"   ***Error : Too many contours %d\n",nsteps);
         recalc = CN_TRUE;
      }
      /* Recalculate the contour step size */
      if (recalc) {
         fprintf(stdout,"   Recalculating the contour step size...\n");
         nsteps = CN_IDLSTEPS;
      }
      cstep  = CNround_to_decimal( (zmax - zmin)/(double)nsteps );
      nsteps = (int) ((zmax-zmin)/cstep);
   }

   /* Ask user for the contour step size */
   while ((nsteps < 1) || (nsteps > CN_MAXSTEPS)) {
      fprintf(stdout,"   Enter contour step size :");
      fscanf(stdin,"%lf",&cstep);
      while ( (c=getc(stdin)) != '\n');
      if (cstep < CN_SMALL)
         fprintf(stdout,"   ***Error : Step size %g is too small\n",cstep);
      else {
         nsteps = (zmax-zmin)/cstep;
         if (nsteps < 1) 
            fprintf(stdout,"   ***Error : Too few contours %d\n",nsteps);
         else if (nsteps > CN_MAXSTEPS) 
            fprintf(stdout,"   ***Error : Too many contours %d\n",nsteps);
      }
   }

   Pptr->cstep  = cstep;
   Pptr->nsteps = nsteps;
   fprintf(stdout,"   Contour step size         = %f\n",cstep);
   fprintf(stdout,"   Number of contours        = %d\n",nsteps);

   /* Double check */
   if (nsteps < 1 || nsteps > CN_MAXSTEPS || cstep < CN_SMALL) {
      fprintf(stderr,"   ***Error - too few/many steps (%d)!\n",nsteps);
      exit(-1);
   }
}


/*
 * Select the contour step size and 
 * cut up the triangles contained in Pptr along the z-planes
 */
void CNslice_contours(Pptr)
CNplotsetptr Pptr;
{
   double level, cstep, zmin, zmax;
   int    lineID = 1;

   /*
    * Select the contour step size
    */
   CNselect_contour_step(Pptr);

   /*
    * get the contour step size and zlimits 
    */
   cstep     = Pptr->cstep;
   zmin      = Pptr->zmin;
   zmax      = Pptr->zmax;

   /*
    * now find the contours 
    */
   level = (int)(zmin/cstep + CN_SMALL)*cstep + cstep;
   while (level <= zmax+CN_SMALL) {
      CNfind_contour(Pptr,level,lineID++);
      level += cstep;
   }
}


/* 
 * find all the segments that are formed when triangle intersects z=level 
 */
void CNfind_contour(Pptr,level,lineID)
CNplotsetptr Pptr;
double       level;
int          lineID;
{
   void       CNdelete_segm();
   CNcurveptr CNinsert_curve();
   CNnodeptr  CNinsert_tailnode();
   static     void find_intsct();
   static     void add_to_tailplot();
   static     void add_to_headplot();

   CNsegmptr  segm_listhead=NULL, segm_listtail=NULL;
   CNsegmptr  S;
   CNtriaptr  T;
   CNcurveptr C;
   CNdatapt   PT1,PT2;
   char       label[CN_MAXCHAR];
   int        linetyp;

   /* 
    * find all the segments of the triangles that intersect a plane
    */
   for (T=Pptr->triahead; T!=NULL; T=T->next)
      find_intsct(level,T,&segm_listhead,&segm_listtail);
   
   /* 
    * The linetype is determined by the order of each slice 
    * The default is 2 linetypes (major and minor)
    */
   if (Pptr->linetypes <=0 || Pptr->linetypes > 3) Pptr->linetypes = 2;
   if (Pptr->linetypes <= 2) 
      linetyp = lineID % Pptr->linetypes;
   else if (Pptr->linetypes == 3) {
      linetyp = lineID % 4;
      if (linetyp == 3) linetyp = 1;
   }
   switch (linetyp) {
   case 0 : linetyp = CN_LNSOLID;    break;
   case 1 : linetyp = CN_LNDOTTED;   break;
   case 2 : linetyp = CN_LNDASHED;   break;
   case 3 : linetyp = CN_LNDOTDASH;  break;
   default: linetyp = CN_LNSOLID;    break;
   }

   /*
    * now sort out the list of segments 
    */
   while ((S=segm_listhead)!=NULL) {
      if (Pptr->zlog) sprintf(label,"%.3g",pow(10.0,level));
      else            sprintf(label,"%.3g",level);
      C = CNinsert_curve(&(Pptr->curvehead), &(Pptr->curvetail),label,linetyp);
      PT1 = S->pt1;
      PT2 = S->pt2;
      CNinsert_tailnode(&PT1, C);
      CNinsert_tailnode(&PT2, C);
      CNdelete_segm(&segm_listhead, &segm_listtail, S);
      add_to_tailplot(&segm_listhead, &segm_listtail, &PT2, C);
      add_to_headplot(&segm_listhead, &segm_listtail, &PT1, C);
   }
}


/* 
 * add to the tail of the plot list 
 */
static void
add_to_tailplot(segm_listhead, segm_listtail, pt, C)
CNsegmptr *segm_listhead, *segm_listtail;
CNdatapt  *pt;
CNcurveptr C;
{
   void      CNdelete_segm();
   CNnodeptr CNinsert_tailnode();
   static    int longline();
   static    CNsegmptr matching_segm();
   CNsegmptr S;

   /* add the further node at every loop iteration */
   while ((S=matching_segm(segm_listhead,pt))!=NULL) {
      if (longline(&(S->pt1),pt) ) {
         CNinsert_tailnode(&(S->pt1),C);
         *pt = S->pt1;
      } else {
         CNinsert_tailnode(&(S->pt2),C);
         *pt = S->pt2;
      }
      CNdelete_segm(segm_listhead, segm_listtail, S);
   }
   return;
}

/* add to the head of the plot list */
static void
add_to_headplot(segm_listhead, segm_listtail, pt, C)
CNsegmptr *segm_listhead, *segm_listtail;
CNdatapt  *pt;
CNcurveptr C;
{
   void      CNdelete_segm();
   CNnodeptr CNinsert_headnode();
   static    int longline();
   static    CNsegmptr matching_segm();
   CNsegmptr S;

   /* add the further node at every loop iteration */
   while ((S=matching_segm(segm_listhead,pt))!=NULL) {
      if (longline(&(S->pt1),pt) ) {
         CNinsert_headnode(&(S->pt1),C);
         *pt = S->pt1;
      } else {
         CNinsert_headnode(&(S->pt2),C);
         *pt = S->pt2;
      }
      CNdelete_segm(segm_listhead, segm_listtail, S);
   }
   return;
}

/* Return segment containing data point that matches a given point */
static CNsegmptr matching_segm(segm_listhead,pt)
CNsegmptr *segm_listhead;
CNdatapt  *pt;
{
   static    int longline();
   CNsegmptr S,SF;
   int       FOUND = CN_FALSE;

   /* check lengths of lines */
   for (S=(*segm_listhead); S!=NULL && !FOUND; S=S->next) {
      /* check the first point */
      if (fabs(S->pt1.x - pt->x) < CN_SMALL) 
         FOUND = !(longline(&(S->pt1),pt));
      if (!FOUND) {
         /* now check the second point */
         if (fabs(S->pt2.x - pt->x) < CN_SMALL) 
            FOUND = !(longline(&(S->pt2),pt));
      }
      if (FOUND) SF = S;
   }
   
   /* return the segment */
   if (!FOUND) SF = NULL;
   return(SF);
}

/* find the segment of intersection between a triangle and a plane */
static void find_intsct(level,T,segm_listhead,segm_listtail)
double    level;
CNtriaptr T;
CNsegmptr *segm_listhead, *segm_listtail;
{
   static int longline();
   static int plane_intsct_tria();
   static int plane_intsct_line();
   CNsegmptr CNinsert_segm();
   CNdatapt  intpts[3],pta;
   int       i=0;

   if (plane_intsct_tria(level,T)) {
      if (plane_intsct_line(level,T->pt1,T->pt2,&pta)) intpts[i++] = pta;
      if (plane_intsct_line(level,T->pt2,T->pt3,&pta)) intpts[i++] = pta;
      if (plane_intsct_line(level,T->pt3,T->pt1,&pta)) intpts[i++] = pta;
      if (i==3) {
         /* only way is to have plane intersect tria at one of the nodes */
         if (longline(&(intpts[0]),&(intpts[1])))
           CNinsert_segm(segm_listhead,segm_listtail,&(intpts[0]),&(intpts[1]));
         else if (longline(&(intpts[1]),&(intpts[2])))
           CNinsert_segm(segm_listhead,segm_listtail,&(intpts[1]),&(intpts[2]));
         else if (longline(&(intpts[2]),&(intpts[0])))
           CNinsert_segm(segm_listhead,segm_listtail,&(intpts[2]),&(intpts[0]));
         else
            /* fprintf(stderr,"Warning: Zero segment length\n")*/;
      } else if (i!=2)
         fprintf(stderr,"Warning: Nonstandard No of intersections = %5d\n",i);
      else if (longline(&(intpts[0]),&(intpts[1])))
         CNinsert_segm(segm_listhead,segm_listtail,&(intpts[0]),&(intpts[1]));
      else
         /*fprintf(stderr,"Warning: Zero segment length\n")*/;
   }
}


/*
 * find out if a line segment is long or short 
 */
static int longline(pta,ptb)
CNdatapt *pta,*ptb;
{
   int    longln = CN_TRUE;
   double dl,dx,dy,dz;

   dx = pta->x - ptb->x;
   dy = pta->y - ptb->y;
   dz = pta->z - ptb->z;
   dl = dx*dx + dy*dy + dz*dz;
   if (dl < CN_SMALLER) longln = CN_FALSE;
   return(longln);
}


/* 
 * find out the intersection point between a line and a z-plane 
 */
static int plane_intsct_line(z,pt1,pt2,pt3)
double z;
CNdatapt pt1,pt2,*pt3;
{
   int    intsct = CN_TRUE;
   double t;

   if ((pt1.z > z && pt2.z > z) || 
       (pt1.z < z && pt2.z < z) ||
       (fabs(pt1.z-pt2.z) < CN_SMALL) )
      intsct = CN_FALSE;
   if (intsct) {
      t = (z - pt1.z)/(pt2.z-pt1.z);
      pt3->x = pt1.x + (pt2.x-pt1.x)*t;
      pt3->y = pt1.y + (pt2.y-pt1.y)*t;
      pt3->z = z;
   }
   return(intsct);
}

/* 
 * find out if the triangle intersects the plane 
 */
static int plane_intsct_tria(z,T)
double z;
CNtriaptr T;
{
   int    intsct = CN_TRUE;
   double z1,z2,z3;

   z1 = T->pt1.z;
   z2 = T->pt2.z;
   z3 = T->pt3.z;
   if ((z1>z && z2>z && z3>z) || (z1<z && z2<z && z3<z))
      intsct = CN_FALSE;
   return(intsct);
}

