/*
 * contread.c - read contour input data from a file 
 * Both of the above routines read in the data in either binary of ascii
 * formats.  If the data is in a rectangular mesh, the mesh is fractured
 * into triangles.  In either case, the triangles are saved in the
 * "CNplotset" data-structure, pointed to by "CNplotsetptr".  The original
 * rectangular mesh is NOT saved.
 */

#include <stdio.h>
#include <math.h>
#include <strings.h>
#include "plot2D.h"

/* 
 * read the data from a file 
 * return NULL upon failure
 */
CNplotsetptr CNread_contourfile(filename, binary, joincurve,
                                cstep, nsteps, linetypes, zlog, ID)
char   *filename;      /* The name of the file                 */
int    binary;         /* Binary/Ascii flag (Ascii=0,Binary=1) */
int    joincurve;      /* Contour-boundary interpolation style */
double cstep;          /* The contour step size                */
int    nsteps;         /* Number of contour steps              */
int    linetypes;      /* No of contour linetypes              */
int    zlog;           /* Logarithmic plot of z-values (log=1) */
int    ID;             /* Plot ID                              */
{
   FILE       *fp;
   CNplotsetptr Pptr;

   /* open the file */
   if (!CNopen_file(filename,&fp)) return(NULL);

   /* Now process the file */
   if (!binary) 
      /* ascii data */
      Pptr = CNread_as_contour(fp,filename,joincurve,
                               cstep,nsteps,linetypes,zlog,ID);
   else
      /* binary data*/
      Pptr = CNread_bn_contour(fp,filename,joincurve,
                               cstep,nsteps,linetypes,zlog,ID);

   /* close the file */
   fclose(fp);

   /* Print out warning message */
   if (Pptr == NULL) 
      fprintf(stderr,"   ***Error! Contour-mesh file-read was unsuccessful.\n");

   /* return */
   return(Pptr);
}


/* 
 * read in the data in ascii format
 * return NULL  upon failure
 */
CNplotsetptr CNread_as_contour(fp, filename, joincurve,
                               cstep, nsteps, linetypes, zlog, ID)
FILE   *fp;            /* File pointer                         */
char   *filename;      /* The name of the file (to be saved)   */
int    joincurve;      /* Contour-boundary interpolation style */
double cstep;          /* The contour step size                */
int    nsteps;         /* Number of contour steps              */
int    linetypes;      /* No of contour linetypes              */
int    zlog;           /* Logarithmic plot of z-values (log=1) */
int    ID;             /* Plot ID                              */
{
   CNplotsetptr CNget_rectilinear_contour_data();
   double      *CNcreate_1D_double_array(); 
 
   CNplotsetptr Pptr = NULL;
   double      *arrptr;
   double       xmin, xmax, ymin, ymax, zmin, zmax;
   double       z, rnx, rny;
   int          nx, ny;
   int          i,j;

   /* Initialize */
   zmin =  CN_LARGE;
   zmax = -CN_LARGE;

   /* read the number of x and y points */
   if (!CNrdfloat(fp,&xmin)) return(NULL);
   if (!CNrdfloat(fp,&xmax)) return(NULL);
   if (!CNrdfloat(fp,&ymin)) return(NULL);
   if (!CNrdfloat(fp,&ymax)) return(NULL);
   if (!CNrdfloat(fp,&rnx )) return(NULL);
   if (!CNrdfloat(fp,&rny )) return(NULL);
   nx   = (int)rnx;
   ny   = (int)rny;

   if ((nx <= 0) || (ny <= 0)) {
      fprintf(stderr,"No of array gridpoints is less than or equal to 0 : ");
      fprintf(stderr,"nx=%d ny=%d\n",nx,ny);
      return(NULL);
   }

   /* allocate space for the data in a 1D array */
   arrptr = CNcreate_1D_double_array(nx*ny);

   /* 
    * read the intensity values and save in a 1D array
    * data goes from i=0 to i=npts-1
    *   a[0,0], a[1,0], a[2,0]...
    */
   for (j=0; j<ny; j++)
   for (i=0; i<nx; i++) {
      if (!CNrdfloat(fp,&z)) return(NULL);   
      if (zlog) {
         if (z < CN_SMALLER) {
            fprintf(stderr,"   ***Error - Z-value (%g) cannot ");
            fprintf(stderr,"be less or equal to zero!\n");
            z = 1.0e-5;
            fprintf(stderr,"      Resetting z to %g\n",z);
         }
         z = log10(z);
      }
      CNset_1D_double_array_value(arrptr,i+j*nx,nx*ny,&z);
      if (z < zmin) zmin = z;
      if (z > zmax) zmax = z;
   }

   /* rearrange the data into a collection of triangles */
   Pptr = CNget_rectilinear_contour_data(filename,
                xmin,xmax,ymin,ymax,zmin,zmax,
                arrptr,nx,ny,joincurve,cstep,nsteps,linetypes,zlog,ID);

   /* free the 1D array */
   CNfree_1D_double_array(arrptr);

   /* return */
   return(Pptr);
}


/* 
 * read in the data in binary format
 * return NULL upon failure
 */
CNplotsetptr CNread_bn_contour(fp, filename, joincurve,
                               cstep, nsteps, linetypes, zlog, ID)
FILE   *fp;            /* File pointer                         */
char   *filename;      /* The name of the file (to be saved)   */
int    joincurve;      /* Contour-boundary interpolation style */
double cstep;          /* The contour step size                */
int    nsteps;         /* Number of contour steps              */
int    linetypes;      /* No of contour linetypes              */
int    zlog;           /* Logarithmic plot of z-values (log=1) */
int    ID;             /* Plot ID                              */
{
   CNplotsetptr CNget_rectilinear_contour_data();
   double      *CNcreate_1D_double_array(); 
   void         free_1D_double_array();

   CNplotsetptr Pptr = NULL;
   double      *arrptr; 

#define         NHEADER  6
   double       xmin, xmax, ymin, ymax, zmin, zmax, z;
   double       header[NHEADER];
   int          nx, ny, npoints;
   int          i, n;

   /* 
    * read header info from the file.
    * read the number of x and y points in the grid
    */
   if ((n = fread((char *)header,sizeof(double),NHEADER,fp)) != NHEADER) {
      fprintf(stderr,"   ***Binary read error of header info!\n");
      exit(1);
   }
   xmin = header[0];
   xmax = header[1];
   ymin = header[2];
   ymax = header[3];
   nx   = (int)header[4];
   ny   = (int)header[5];

   /* Check the number of grid points */
   if ((nx <= 0) || (ny <= 0)) {
      fprintf(stderr,"No of array gridpoints is less than or equal to 0 : ");
      fprintf(stderr,"nx=%d ny=%d\n",nx,ny);
      return(NULL);
   }

   /* allocate space for the data in a 1D array */
   npoints = nx*ny;
   arrptr  = CNcreate_1D_double_array(npoints);

   /* 
    * read the intensity values and save in a 1D array
    * data goes from i=0 to i=npts-1
    */
   if ((n = fread((char *)arrptr,sizeof(double),npoints,fp)) != npoints) {
      fprintf(stderr,"   ***Binary read error of data array!\n");
      exit(1);
   }

   /*
    * Take the log of the z-value if necessary
    */
   if (zlog) {
      for (i=0; i<nx*ny; i++) {
         z = CNget_1D_double_array_value(arrptr,i,nx*ny);
         if (z < CN_SMALLER) {
            fprintf(stderr,"   ***Error - Z-value (%g) cannot ");
            fprintf(stderr,"be less or equal to zero!\n");
            z = 1.0e-5;
            fprintf(stderr,"      Resetting z to %g\n",z);
         }
         z = log10(z);
         CNset_1D_double_array_value(arrptr,i,nx*ny,&z);
      }
   }

   /*
    * find the z-min and max values
    */
   CNget_1D_double_array_maxmin(arrptr,nx*ny,&zmin,&zmax);

   /* rearrange the data into a collection of triangles */
   Pptr = CNget_rectilinear_contour_data(filename,
                xmin,xmax,ymin,ymax,zmin,zmax,
                arrptr,nx,ny,joincurve,cstep,nsteps,linetypes,zlog,ID);

   /* free the 1D array */
   CNfree_1D_double_array(arrptr);

   /* return */
   return(Pptr);
}

/* 
 * read the data from a file in 3D format (list of triangles)
 * return NULL upon failure
 */
CNplotsetptr CNread_3Dcontourfile(filename, binary,
                                  cstep, nsteps, linetypes, zlog, ID)
char   *filename;      /* The name of the file                 */
int    binary;         /* Binary/Ascii flag (Ascii=0,Binary=1) */
double cstep;          /* The contour step size                */
int    nsteps;         /* Number of contour steps              */
int    linetypes;      /* No of contour linetypes              */
int    zlog;           /* Logarithmic plot of z-values (log=1) */
int    ID;             /* Plot ID                              */
{
   FILE       *fp;
   CNplotsetptr Pptr;

   /* open the file */
   if (!CNopen_file(filename,&fp)) return(NULL);

   /* Now process the file */
   if (!binary) 
      /* ascii data */
      Pptr = CNread_as_3Dcontour(fp,filename,cstep,nsteps,linetypes,zlog,ID); 
   else 
      /* binary data */
      Pptr = CNread_bn_3Dcontour(fp,filename,cstep,nsteps,linetypes,zlog,ID); 

   /* close the file */
   fclose(fp);

   /* Print out warning message */
   if (Pptr == NULL) 
      fprintf(stderr,"   ***Error! 3D file-read was unsuccessful.\n");

   /* return */
   return(Pptr);
}

/* 
 * read in the data in ascii 3D format
 * return NULL  upon failure
 */
CNplotsetptr CNread_as_3Dcontour(fp,filename,cstep,nsteps,linetypes,zlog,ID)
FILE   *fp;            /* File pointer                         */
char   *filename;      /* The name of the file (to be saved)   */
double cstep;          /* The contour step size                */
int    nsteps;         /* Number of contour steps              */
int    linetypes;      /* No of contour linetypes              */
int    zlog;           /* Logarithmic plot of z-values (log=1) */
int    ID;             /* Plot ID                              */
{
   CNplotsetptr CNget_triangular_contour_data();
   CNtriaptr    CNinsert_tria();
 
   CNplotsetptr Pptr = NULL;
   CNdatapt     point[4];
   CNtriaptr    triahead = NULL, triatail = NULL;
   double       xmin, xmax, ymin, ymax, zmin, zmax;
   double       z, rncv;
   int          ncurves, npts;
   int          i,m;

   /* read the boundary info and no of triangles */
   if (!CNrdfloat(fp,&xmin)) return(NULL);
   if (!CNrdfloat(fp,&xmax)) return(NULL);
   if (!CNrdfloat(fp,&ymin)) return(NULL);
   if (!CNrdfloat(fp,&ymax)) return(NULL);
   if (!CNrdfloat(fp,&zmin)) return(NULL);
   if (!CNrdfloat(fp,&zmax)) return(NULL);
   if (!CNrdfloat(fp,&rncv)) return(NULL);

   /* No of triangles */
   ncurves = (int)rncv;
   if (ncurves <= 0) {
      fprintf(stderr,"   ***Error: Plot Curves (%d) <= 0\n",ncurves);
      return(NULL);
   }
   fprintf(stdout,"   Reading %d triangles from 3D dataset...\n",ncurves);

   /* Reinitialize */
   zmin =  CN_LARGE;
   zmax = -CN_LARGE;

   for (m=0; m<ncurves; m++) {
      if (!CNrdfloat(fp,&rncv)) return(NULL);
      npts = (int)rncv;
      if (npts <= 0) {
         fprintf(stderr,"   ***Error: Plot Points (%d) <= 0\n",npts);
         return(NULL);
      } else if (npts != 3 && npts != 4) {
         fprintf(stderr,"   ***Error! 3D datasets can only consist ");
         fprintf(stderr,"of 3D Triangles!\n");
         fprintf(stderr,"      Cannot read %d-sided polygon...\n",npts);
         return(NULL);
      } else {
         for (i=0; i<npts; i++) {
            if (!CNrdfloat(fp,&z)) return(NULL); point[i].x = z; 
            if (!CNrdfloat(fp,&z)) return(NULL); point[i].y = z; 
            if (!CNrdfloat(fp,&z)) return(NULL); point[i].z = z; 
            if (zlog) {
               if (point[i].z < CN_SMALLER) {
                  fprintf(stderr,"   ***Error - Z-value (%g) cannot ");
                  fprintf(stderr,"be less or equal to zero!\n");
                  point[i].z = 1.0e-5;
                  fprintf(stderr,"      Resetting z to %g\n",point[i].z);
               }
               point[i].z = log10(point[i].z);
            }
            if (point[i].z > zmax) zmax = point[i].z;
            if (point[i].z < zmin) zmin = point[i].z;
         }
         CNinsert_tria(&triahead,&triatail,&(point[0]),&(point[1]),&(point[2]));
      }
   }
   fprintf(stdout,"   Finished reading %d triangles from 3D dataset\n",ncurves);
   fflush(stdout);

   /* slice and dice the triangles and place results in a linked list */
   Pptr = CNget_triangular_contour_data(filename,
                xmin,xmax,ymin,ymax,zmin,zmax,
                triahead,triatail,cstep,nsteps,linetypes,zlog,ID);

   /* return */
   return(Pptr);
}


/* 
 * read in the data in binary format
 * return NULL upon failure
 */
/*ARGSUSED*/
CNplotsetptr CNread_bn_3Dcontour(fp,filename,cstep,nsteps,linetypes,zlog,ID)
FILE   *fp;            /* File pointer                         */
char   *filename;      /* The name of the file (to be saved)   */
double cstep;          /* The contour step size                */
int    nsteps;         /* Number of contour steps              */
int    linetypes;      /* No of contour linetypes              */
int    zlog;           /* Logarithmic plot of z-values (log=1) */
int    ID;             /* Plot ID                              */
{
   fprintf(stderr,"Sorry - cannt read 3D binary data yet!\n");
   return(NULL);
}

