/*** list.c - linked list procedures for 2D plots (x-y, contour) ***/

#include <stdio.h>
#include <strings.h>
#include "plot2D.h"

/*
 * TRIANGLES
 *    The triangle is the basic mesh element for a 3D surface.
 *    A triangle contains 3 points/vertices
 */


/*
 * triangles are used to store surface information 
 */
static
CNtriaptr make_tria(p1,p2,p3)
CNdatapt *p1,*p2,*p3;
{
   char *malloc();
   CNtriaptr newptr;

   if ((newptr = (CNtriaptr)malloc(sizeof(CNtria)))!=NULL) {
      newptr->pt1 = *p1;
      newptr->pt2 = *p2;
      newptr->pt3 = *p3;
      newptr->next = NULL;
      newptr->prev = NULL;
   }
   return(newptr);
}


/* 
 * Insert a tria at the tail of the current triangle list 
 */
CNtriaptr CNinsert_tria(tria_listhead,tria_listtail,p1,p2,p3)
CNtriaptr *tria_listhead, *tria_listtail;
CNdatapt *p1,*p2,*p3;
{
   static  CNtriaptr make_tria();
   CNtriaptr prev,next,A,B;

   A = *tria_listtail;
   if ((B=make_tria(p1,p2,p3))!=NULL) {
      if (A==NULL) {
         *tria_listhead = B;
         *tria_listtail = B;
      } else {
         prev = A->prev;
         next = A->next;
         B->next = next;
         B->prev = A;
         A->next = B;
         if (next    != NULL) next->prev = B;
         if (B->next == NULL) *tria_listtail = B;
      }
   }
   return(B);
}


/* 
 * print out the list of triangles 
 */
/*ARGSUSED*/
void CNprint_tria_list(tria_listhead, tria_listtail)
CNtriaptr tria_listhead, tria_listtail;
{
   static void print_point();
   CNtriaptr T;

   for (T=tria_listhead; T!=NULL; T=T->next) {
      print_point(&(T->pt1));
      print_point(&(T->pt2));
      print_point(&(T->pt3));
      fprintf(stdout,"\n");
      fflush(stdout);
   }
}


/* 
 * Delete triaptr 
 */
void CNdelete_triaptr(tria_listhead, tria_listtail, T)
CNtriaptr *tria_listhead, *tria_listtail;
CNtriaptr T;
{
   CNtriaptr prev,next;

   prev = T->prev;
   next = T->next;
   if (prev!=NULL) prev->next = next;
   if (next!=NULL) next->prev = prev;
   if (T==*tria_listhead) *tria_listhead = next;
   if (T==*tria_listtail) *tria_listtail = prev;

   /* Now delete T */
   free ((char*)T);
}


/* 
 * print the coordinates of a point/vertice
 */
static void print_point(pt)
CNdatapt *pt;
{
   fprintf(stdout,"x= %10.5f y= %10.5f z= %10.5f\n",pt->x,pt->y,pt->z);
}



/*
 * SEGMENTS
 *    A segment contains 2 points/vertices 
 *    The contouring works by slicing up a list of triangles, which 
 *    generates a list of segments.  The segments are attached to each
 *    other to form a nodelist.
 */


/*
 * Allocate room for a segment 
 */
static CNsegmptr make_segm(p1,p2)
CNdatapt *p1,*p2;
{
   char *malloc();
   CNsegmptr newptr;

   if ((newptr = (CNsegmptr)malloc(sizeof(CNsegm)))!=NULL) {
      newptr->pt1 = *p1;
      newptr->pt2 = *p2;
      newptr->next = NULL;
      newptr->prev = NULL;
   }
   return(newptr);
}


/*  
 * Insert a segm at the tail of the current segment list
 */
CNsegmptr CNinsert_segm(segm_listhead, segm_listtail, p1, p2)
CNsegmptr *segm_listhead, *segm_listtail;
CNdatapt *p1,*p2;
{
   static CNsegmptr make_segm();
   CNsegmptr prev,next,A,B;

   A = *segm_listtail;
   if ((B=make_segm(p1,p2))!=NULL) {
      if (A==NULL) {
         *segm_listhead = B;
         *segm_listtail = B;
      } else {
         prev = A->prev;
         next = A->next;
         B->next = next;
         B->prev = A;
         A->next = B;
         if (next != NULL) next->prev = B;
         if (B->next == NULL) *segm_listtail = B;
      }
   }
   return(B);
}


/*
 * Delete segm at address L 
 */
void CNdelete_segm(segm_listhead, segm_listtail, L)
CNsegmptr *segm_listhead, *segm_listtail;
CNsegmptr L;
{
   CNsegmptr prev,next;

   prev = L->prev;
   next = L->next;
   if (prev!=NULL) prev->next = next;
   if (next!=NULL) next->prev = prev;
   if (L== *segm_listhead) *segm_listhead = next;
   if (L== *segm_listtail) *segm_listtail = prev;
   free ((char*)L);
   L = NULL;
}


/* 
 * Print out the list of segments 
 */
/*ARGSUSED*/
void CNprint_segm_list(segm_listhead, segm_listtail)
CNsegmptr segm_listhead, segm_listtail;
{
   static void print_point();
   CNsegmptr S;

   for (S=segm_listhead; S!=NULL; S=S->next) {
      print_point(&(S->pt1));
      print_point(&(S->pt2));
      fprintf(stdout,"\n");
      fflush(stdout);
   }
}



/*
 * NODE DATA STRUCTURE
 *    nodes are used to store all the points of the separate 
 *    curves/contours 
 */


/* 
 * Allocate room for a node 
 */
static CNnodeptr make_node(p1)
CNdatapt *p1;
{
   char *malloc();
   CNnodeptr newptr;

   if ((newptr = (CNnodeptr)malloc(sizeof(CNnode)))!=NULL) {
      newptr->pt   = *p1;
      newptr->next = NULL;
      newptr->prev = NULL;
   }
   return(newptr);
}


/* 
 * Insert a node at the tail of the current curve list 
 */
CNnodeptr CNinsert_tailnode(p1,C)
CNdatapt *p1;
CNcurveptr C;
{
   static  CNnodeptr make_node();
   CNnodeptr prev,next,A,B;

   A = C->nodetail;
   if ((B=make_node(p1))!=NULL) {
      if (A==NULL) {
         C->nodehead = B;
         C->nodetail = B;
      } else {
         prev = A->prev;
         next = A->next;
         B->next = next;
         B->prev = A;
         A->next = B;
         if (next    != NULL) next->prev = B;
         if (B->next == NULL) C->nodetail = B;
      }
   }
   return(B);
}


/*
 * Insert a node at the head of the current curve list
 */
CNnodeptr CNinsert_headnode(p1,C)
CNdatapt *p1;
CNcurveptr C;
{
   static CNnodeptr make_node();
   CNnodeptr prev,next,A,B;

   A = C->nodehead;
   if ((B=make_node(p1))!=NULL) {
      if (A==NULL) {
         C->nodehead = B;
         C->nodetail = B;
      } else {
         B->next = A;
         A->prev = B;
         C->nodehead = B;
      }
   }
   return(B);
}


/* 
 * Delete node on curve C at address A 
 */
void CNdelete_node(A,C)
CNnodeptr A;
CNcurveptr C;
{
   CNnodeptr prev,next;

   prev = A->prev;
   next = A->next;
   if (prev!=NULL) prev->next = next;
   if (next!=NULL) next->prev = prev;
   if (A==C->nodehead) C->nodehead = next;
   if (A==C->nodetail) C->nodetail = prev;
   free ((char*)A);
   A = NULL;
}


/*
 * Print the list of nodes in a curve 
 */
void CNprint_node_list(C)
CNcurveptr C;
{
   static void print_point();
   CNnodeptr N;

   for (N=C->nodehead; N!=NULL; N=N->next) 
      print_point(&(N->pt));
   fprintf(stdout,"\n");
   fflush(stdout);
}



/*
 * CURVE PLOT DATA STRUCTURE
 *    curves are used to store the nodes of each separate contour or
 *    plot-curve 
 */

/* Allocate room for a curve */
static CNcurveptr make_curve(label,linetype)
char *label;
int   linetype;
{
   char *malloc();
   CNcurveptr newptr;

   if ((newptr = (CNcurveptr)malloc(sizeof(CNcurve)))!=NULL) {
      strcpy(newptr->label, label);
      newptr->linewidth  = 1;
      newptr->linetype   = linetype;
      newptr->markertype = 1;
      newptr->nodehead   = NULL;
      newptr->nodetail   = NULL;
      newptr->next       = NULL;
      newptr->prev       = NULL;
   }
   return(newptr);
}

/* Insert a curve at the tail of the current curve list */
CNcurveptr CNinsert_curve(curve_listhead, curve_listtail, label, linetype)
CNcurveptr *curve_listhead, *curve_listtail;
char       *label;
int        linetype;
{
   static CNcurveptr make_curve();
   CNcurveptr prev,next,A,B;

   A = *curve_listtail;
   if ((B=make_curve(label,linetype))!=NULL) {
      if (A==NULL) {
         *curve_listhead = B;
         *curve_listtail = B;
      } else {
         prev = A->prev;
         next = A->next;
         B->next = next;
         B->prev = A;
         A->next = B;
         if (next != NULL) next->prev = B;
         if (B->next == NULL) *curve_listtail = B;
      }
   }
   return(B);
}


/* 
 * print the curves 
 */
/*ARGSUSED*/
void CNprint_curve_list(curve_listhead, curve_listtail)
CNcurveptr curve_listhead, curve_listtail;
{
   void     CNprint_node_list();
   CNcurveptr C;

   for (C=curve_listhead; C!=NULL; C=C->next) {
      fprintf(stdout,"Label = %s   Linetype = %d\n",C->label,C->linetype);
      CNprint_node_list(C);
   }
}

 
/*
 * Delete curveptr
 */
void CNdelete_curveptr(curve_listhead, curve_listtail, P)
CNcurveptr *curve_listhead, *curve_listtail;
CNcurveptr P;
{
   void     CNdelete_node();
   CNcurveptr prev,next;

   prev = P->prev;
   next = P->next;
   if (prev!=NULL) prev->next = next;
   if (next!=NULL) next->prev = prev;
   if (P==*curve_listhead) *curve_listhead = next;
   if (P==*curve_listtail) *curve_listtail = prev;

   /* delete all the nodes in P */
   while (P->nodehead !=NULL)
       CNdelete_node(P->nodehead,P);

   /* Now delete P */
   free ((char*)P);
}



/*
 * PLOT DATA STRUCTURE
 *    A plotset contains a set of curves and additional data related to
 *    the plot.
 */


/* 
 * Allocate room for the plotset 
 */
CNplotsetptr CNmake_plotset(filename,plottype,
                            xmin,xmax,ymin,ymax,zmin,zmax,
                            cstep,nsteps,linetypes,zlog,ID)
char   *filename;
double xmin,xmax,ymin,ymax,zmin,zmax,cstep;
int    plottype,nsteps,linetypes,zlog,ID;
{
   char *malloc();
   CNplotsetptr newptr;

   if ((newptr = (CNplotsetptr)malloc(sizeof(CNplotset)))!=NULL) {
      strcpy(newptr->filename,filename);
      newptr->plottype = plottype;
      newptr->xmin     = xmin;
      newptr->xmax     = xmax;
      newptr->ymin     = ymin;
      newptr->ymax     = ymax;
      newptr->zmin     = zmin;
      newptr->zmax     = zmax;
      newptr->cstep    = cstep;
      newptr->nsteps   = nsteps;
      newptr->linetypes= linetypes;
      newptr->zlog     = zlog;
      newptr->ID       = ID;
      newptr->triahead = NULL;
      newptr->triatail = NULL;
      newptr->curvehead= NULL;
      newptr->curvetail= NULL;
      newptr->next     = NULL;
      newptr->prev     = NULL;
   }
   return(newptr);
}


/* 
 * Insert a plotset at the tail of the current list 
 * Return 0 if failed.
 */
int CNinsert_plotset(plot_listhead, plot_listtail, Pptr)
CNplotsetptr *plot_listhead, *plot_listtail;
CNplotsetptr Pptr;
{
   CNplotsetptr prev,next,A;

   A = *plot_listtail;
   if (Pptr != NULL) {
      if (A==NULL) {
         *plot_listhead = Pptr;
         *plot_listtail = Pptr;
      } else {
         prev = A->prev;
         next = A->next;
         Pptr->next = next;
         Pptr->prev = A;
         A->next    = Pptr;
         if (next != NULL) next->prev = Pptr;
         if (Pptr->next == NULL) *plot_listtail = Pptr;
      }
      return(1);
   } else {
     fprintf(stderr,"   ***Error! Null pointer passed to insert_plotset()\n");
     return(0);
   }
}


/*
 * print the data in the plot set 
 */
/*ARGSUSED*/
void CNprint_plotset_list(plot_listhead, plot_listtail)
CNplotsetptr plot_listhead, plot_listtail;
{
#define PLOTTYPE(x) ((x == CN_CONTOUR) ? "Contour" : "2D Plot")

   void       CNprint_curve_list();
   CNplotsetptr Pptr;
  
   for (Pptr=plot_listhead; Pptr!=NULL; Pptr=Pptr->next) {
      fprintf(stdout,"\n");
      fprintf(stdout,"PLOT ID   = %d\n",Pptr->ID);
      fprintf(stdout,"PLOT TYPE = %s\n",PLOTTYPE(Pptr->plottype));
      fprintf(stdout,"Origin    = %s\n",Pptr->filename);
      if (Pptr->plottype == CN_CONTOUR) {
         fprintf(stdout,"Linetypes = %d\n",Pptr->linetypes);
         fprintf(stdout,"Cstep     = %f\n",Pptr->cstep);
         fprintf(stdout,"Nsteps    = %d\n",Pptr->nsteps);
         fprintf(stdout,"Z-Log     = %s\n",((Pptr->zlog) ? "TRUE" : "FALSE"));
      }
      fprintf(stdout,"xmin = %7.3g   xmax = %7.3g\n",Pptr->xmin,Pptr->xmax);
      fprintf(stdout,"ymin = %7.3g   ymax = %7.3g\n",Pptr->ymin,Pptr->ymax);
      fprintf(stdout,"zmin = %7.3g   zmax = %7.3g\n",Pptr->zmin,Pptr->zmax);
      CNprint_curve_list(Pptr->curvehead, Pptr->curvetail);
   }
} 

/*
 * delete the contour set 
 */
void CNdelete_plotset(plot_listhead, plot_listtail, Pptr)
CNplotsetptr *plot_listhead, *plot_listtail;
CNplotsetptr Pptr;
{
   void       CNdelete_curveptr();
   void       CNdelete_triaptr();
   CNplotsetptr prev,next;

   prev = Pptr->prev;
   next = Pptr->next;
   if (prev!=NULL) prev->next = next;
   if (next!=NULL) next->prev = prev;
   if (Pptr==*plot_listhead) *plot_listhead = next;
   if (Pptr==*plot_listtail) *plot_listtail = prev;

   /* delete all the curveptrs in P */
   while (Pptr->curvehead !=NULL)
      CNdelete_curveptr(&(Pptr->curvehead),&(Pptr->curvetail),Pptr->curvehead);

   /* delete all the triangles in C */
   while (Pptr->triahead !=NULL)
      CNdelete_triaptr(&(Pptr->triahead),&(Pptr->triatail),Pptr->triahead);

   /* Now delete Pptr */
   free ((char*)Pptr);
}


/* delete the entire list of plotsets */
void CNdelete_plotset_list(plot_listhead, plot_listtail)
CNplotsetptr *plot_listhead, *plot_listtail;
{ 
   void CNdelete_plotset();

   while (*plot_listhead != NULL) 
      CNdelete_plotset(plot_listhead, plot_listtail, *plot_listhead);
}


/* 
 * Store the linked list in the contour data structure 
 */
void CNstore_plotset(plothead, plottail, pptr, verbose)
CNplotsetptr *plothead, *plottail, pptr;
int          verbose;
{
   if (pptr == NULL) {
      /* Read error - print out warning message */
      fprintf(stderr, "   ***Error during data read!\n");
   } else if (!CNinsert_plotset(plothead, plottail, pptr)) {
      fprintf(stderr,"   ***Error: Couldn't add pointer to list\n");
      exit(-1);
   } else {
      /* Things are going well here */
      if (verbose)
         CNprint_plotset_list(*plothead, *plottail);
   }
}
