/*
 * contour.h - definitions for the contour program
 */

/*
 * Constants 
 */
#define CN_MAXCHAR 1000          /* No of characters in a word */
#define CN_TRUE    1             /* True                       */
#define CN_FALSE   0             /* False                      */

#define CN_LARGE   1.0e10        /* A large number             */
#define CN_SMALL   1.0e-10       /* A small number             */
#define CN_SMALLER 1.0e-30       /* A very small number        */

#define CN_HIGH    2             /* Contour-boundary interpolation style */
#define CN_LOW     1             /* Contour-boundary interpolation style */
#define CN_NONE    0             /* Contour-boundary interpolation style */

#define CN_PLOT2D  1             /* 2D plot                    */
#define CN_CONTOUR 2             /* contour plot               */

#define CN_LNSOLID   1           /* Line type                  */
#define CN_LNDASHED  2           /* Line type                  */
#define CN_LNDOTTED  3           /* Line type                  */
#define CN_LNDOTDASH 4           /* Line type                  */

/*
 * General plot data structure - the basic mesh element is a triangle
 *    A triangle consists of 3 points
 *    A segment is contains 2 points 
 *    A node is a point but with linked list info attached.
 *
 *    A curve is made up of a linked list of nodes
 *    A plotset consists of a linked list of curves
 */

/* x-y-z coordinates */
typedef struct data_strct {
   double x;
   double y;
   double z;
} CNdatapt;

/* triangles */
typedef struct CNtria_strct {
   CNdatapt pt1;
   CNdatapt pt2;
   CNdatapt pt3;
   struct CNtria_strct *next;
   struct CNtria_strct *prev;
} CNtria;

typedef struct CNtria_strct *CNtriaptr;

/* segments */
typedef struct CNsegm_strct {
   CNdatapt pt1;
   CNdatapt pt2;
   struct CNsegm_strct *next;
   struct CNsegm_strct *prev;
} CNsegm;

typedef struct CNsegm_strct *CNsegmptr;

/* nodes */
typedef struct CNnode_strct {
   CNdatapt pt;
   struct CNnode_strct *next;
   struct CNnode_strct *prev;
} CNnode;

typedef struct CNnode_strct *CNnodeptr;

/* curves */
typedef struct CNcurve_strct {
   short  linewidth;
   short  linetype;
   short  markertype;
   char   label[CN_MAXCHAR];
   struct CNnode_strct *nodehead;
   struct CNnode_strct *nodetail;
   struct CNcurve_strct *next;
   struct CNcurve_strct *prev;
} CNcurve;

typedef struct CNcurve_strct *CNcurveptr;

/* plot sets */
typedef struct CNplotset_strct {
   short  ID;
   short  plottype;
   double xmin;
   double xmax;
   double ymin;
   double ymax;
   double zmin;
   double zmax;
   double cstep;
   short  nsteps;
   short  linetypes;
   short  zlog;
   char   filename[CN_MAXCHAR];
   struct CNtria_strct *triahead;
   struct CNtria_strct *triatail;
   struct CNcurve_strct *curvehead;
   struct CNcurve_strct *curvetail;
   struct CNplotset_strct *next;
   struct CNplotset_strct *prev;
} CNplotset;

typedef struct CNplotset_strct *CNplotsetptr;



/*
 *  procedures
 */

/* 1D array */
extern double      *CNcreate_1D_double_array();
extern void         CNset_1D_double_array_value();
extern double       CNget_1D_double_array_value();
extern void         CNfree_1D_double_array();
extern void         CNget_1D_double_array_maxmin();
extern CNdatapt    *CNcreate_data_array();
extern void         CNset_data_array_value();
extern CNdatapt     CNget_data_array_value();
extern void         CNfree_data_array();

/* Linked lists */
extern CNtriaptr    CNinsert_tria();
extern void         CNprint_tria_list();
extern void         CNdelete_triaptr();
extern CNsegmptr    CNinsert_segm();
extern void         CNdelete_segm();
extern void         CNprint_segm_list();
extern CNnodeptr    CNinsert_tailnode();
extern CNnodeptr    CNinsert_headnode();
extern void         CNdelete_node();
extern void         CNprint_node_list();
extern CNcurveptr   CNinsert_curve();
extern void         CNprint_curve_list();
extern void         CNdelete_curveptr();
extern CNplotsetptr CNmake_plotset();
extern int          CNinsert_plotset();
extern void         CNprint_plotset_list();
extern void         CNdelete_plotset_list();
extern void         CNstore_plotset();

/* Rounding */
extern double CNround_to_decimal();

/* Read generic data */
extern int          CNrdfloat();
extern int          CNopen_file();

/* Read from contour files */
extern CNplotsetptr CNread_contourfile();
extern CNplotsetptr CNread_as_contour();
extern CNplotsetptr CNread_bn_contour();
extern CNplotsetptr CNread_3Dcontourfile();
extern CNplotsetptr CNread_as_3Dcontour();
extern CNplotsetptr CNread_bn_3Dcontour();

/* Read from plot2D files */
extern CNplotsetptr CNread_plot2Dfile();
extern CNplotsetptr CNread_as_plot2D();
extern CNplotsetptr CNread_bn_plot2D();

/* Manipulate the contour data */
extern CNplotsetptr CNget_rectilinear_contour_data();
extern void         CNselect_contour_step();
extern void         CNslice_contours();
extern void         CNfind_contour();
extern void         CNset_contour_boundaries();
extern void         CNexpand_cont();
extern void         CNexpand_mesh();
extern void         CNcalc_single_cont();

/* String operations */
extern void         CNstring_to_upper();
extern void         CNstring_to_lower();

/* Misc Functions */
extern void         CNget_localtime();
extern double       CNround_to_decimal();
extern void         CNget_autorange();
extern double       CNauto_round_down();
extern double       CNauto_round_up();
extern double       CNconv_to_log();
