/*
 * plot2d_read.c - read 2D drawplot input data from a file 
 * Both of the above routines read in the data in either binary of ascii
 * formats.  The data is saved in the
 * "CNplotset" data-structure, pointed to by "CNplotsetptr".  
 */

#include <stdio.h>
#include <math.h>
#include <strings.h>
#include "plot2D.h"

/* 
 * read the data from a file 
 * return NULL upon failure
 */
CNplotsetptr CNread_plot2Dfile(filename, binary, 
                               xlog, ylog, ID)
char   *filename;      /* The name of the file                 */
int    binary;         /* Binary/Ascii flag (Ascii=0,Binary=1) */
int    xlog;           /* Logarithmic plot of x-values (log=1) */
int    ylog;           /* Logarithmic plot of y-values (log=1) */
int    ID;             /* Plot ID                              */
{
   FILE         *fp;
   CNplotsetptr Pptr;

   /* open the file */
   if (!CNopen_file(filename,&fp)) return(NULL);

   /* Now process the file */
   if (!binary) 
      /* ascii data */
      Pptr = CNread_as_plot2D(fp,filename,xlog,ylog,ID);
   else
      /* binary data*/
      Pptr = CNread_bn_plot2D(fp,filename,xlog,ylog,ID);

   /* close the file */
   fclose(fp);

   /* Print out warning message */
   if (Pptr == NULL) 
      fprintf(stderr,"   ***Error! Plot2D file-read was unsuccessful.\n");

   /* return */
   return(Pptr);
}


/* 
 * read in the data in ascii format
 * return NULL  upon failure
 */
CNplotsetptr CNread_as_plot2D(fp, filename, xlog, ylog, ID)
FILE   *fp;            /* File pointer                         */
char   *filename;      /* The name of the file (to be saved)   */
int    xlog;           /* Logarithmic plot of x-values (log=1) */
int    ylog;           /* Logarithmic plot of y-values (log=1) */
int    ID;             /* Plot ID                              */
{
   CNplotsetptr Pptr = NULL;
   CNcurveptr   curvehead = NULL, curvetail = NULL, Cptr = NULL;
   CNdatapt     pt;
   double       xmin, xmax, ymin, ymax, zmin=0.0, zmax=0.0;
   double       rncv, rnpts, x, y;
   int          ncv, npts;
   int          i,m;

   /* read the number of x and y points */
   if (!CNrdfloat(fp,&xmin)) return(NULL);
   if (!CNrdfloat(fp,&xmax)) return(NULL);
   if (!CNrdfloat(fp,&ymin)) return(NULL);
   if (!CNrdfloat(fp,&ymax)) return(NULL);
   if (!CNrdfloat(fp,&rncv)) return(NULL);
   ncv  = (int)rncv;

   /* Check number of curves */
   if (ncv <= 0) {
      fprintf(stderr,"No of plot curves is less than or equal to 0 : ");
      fprintf(stderr,"ncurves=%d\n",ncv);
      return(NULL);
   }

   /* Convert to log axes if necessary */
   if (xlog) {
      xmin = CNauto_round_down(xmin);
      xmax = CNauto_round_up  (xmax);
      xmin = CNconv_to_log(xmin);
      xmax = CNconv_to_log(xmax);
   }
   if (ylog) {
      ymin = CNauto_round_down(ymin);
      ymax = CNauto_round_up  (ymax);
      ymin = CNconv_to_log(ymin);
      ymax = CNconv_to_log(ymax);
   }

   /* Read and store the curves */
   for (m=0; m<ncv; m++) {
      /* Allocate a curve data structure */
      Cptr = CNinsert_curve(&curvehead, &curvetail, "2D", 1);

      /* Read the number of points */
      if (!CNrdfloat(fp,&rnpts)) return(NULL);
      npts = (int)rnpts;

      /* Check number of points */
      if (npts <= 0) {
         fprintf(stderr,"No of plot points is less than or equal to 0 : ");
         fprintf(stderr,"npoints=%d\n",npts);
         return(NULL);
      }

      /* Read curve points */
      for (i=0; i<npts; i++) {
         if (!CNrdfloat(fp,&x)) return(NULL);   
         if (!CNrdfloat(fp,&y)) return(NULL);   
         if (xlog) x = CNconv_to_log(x);
         if (ylog) y = CNconv_to_log(y);
         pt.x = x;
         pt.y = y;
         pt.z = 0.0;
         CNinsert_tailnode(&pt, Cptr);
      }
   }

   /* Store the entire dataset in a plotset */
   Pptr = CNmake_plotset(filename,CN_PLOT2D,
                xmin,xmax,ymin,ymax,zmin,zmax,
                0.0,0,1,0,ID);
   Pptr->curvehead = curvehead;
   Pptr->curvetail = curvetail;

   /* return */
   return(Pptr);
}


/* 
 * read in the data in binary format
 * return NULL upon failure
 */
CNplotsetptr CNread_bn_plot2D(fp, filename, xlog, ylog, ID)
FILE   *fp;            /* File pointer                         */
char   *filename;      /* The name of the file (to be saved)   */
int    xlog;           /* Logarithmic plot of x-values (log=1) */
int    ylog;           /* Logarithmic plot of y-values (log=1) */
int    ID;             /* Plot ID                              */
{
#define         NHEADER  5
   CNplotsetptr Pptr = NULL;
   CNcurveptr   curvehead = NULL, curvetail = NULL, Cptr = NULL;
   CNdatapt     pt;
   double       xmin, xmax, ymin, ymax, zmin=0.0, zmax=0.0;
   double       header[NHEADER];
   double       *arrptr;
   int          ncv, npts;
   int          i, m, n;

   /* 
    * read header info from the file.
    * read the number of x and y points in the grid
    */
   if ((n = fread((char *)header,sizeof(double),NHEADER,fp)) != NHEADER) {
      fprintf(stderr,"   ***Binary read error of header info!\n");
      exit(1);
   }
   xmin = header[0];
   xmax = header[1];
   ymin = header[2];
   ymax = header[3];
   ncv  = (int)header[4];

   /* Check number of curves */
   if (ncv <= 0) {
      fprintf(stderr,"No of plot curves is less than or equal to 0 : ");
      fprintf(stderr,"ncurves=%d\n",ncv);
      return(NULL);
   }

   /* Convert to log axes if necessary */
   if (xlog) {
      xmin = CNauto_round_down(xmin);
      xmax = CNauto_round_up  (xmax);
      xmin = CNconv_to_log(xmin);
      xmax = CNconv_to_log(xmax);
   }
   if (ylog) {
      ymin = CNauto_round_down(ymin);
      ymax = CNauto_round_up  (ymax);
      ymin = CNconv_to_log(ymin);
      ymax = CNconv_to_log(ymax);
   }

   /*
    * Read the points and store them
    */
   for (m=0; m<ncv; m++) {
      /* Allocate a curve data structure */
      Cptr = CNinsert_curve(&curvehead, &curvetail, "2D", 1);

      /* Read the number of points */
      if ((n = fread((char *)header,sizeof(double),1,fp)) != 1) {
         fprintf(stderr,"   ***Binary read error of header info (npts)!\n");
         exit(1);
      }

      /* Check number of points */
      npts = (int)header[0];
      if (npts <= 0) {
         fprintf(stderr,"No of plot points is less than or equal to 0 : ");
         fprintf(stderr,"npoints=%d\n",npts);
         return(NULL);
      }

      /* allocate space for the data in a 1D array */
      arrptr  = CNcreate_1D_double_array(npts*2);

      /* Read the x-y values and store them in a temp array */
      if ((n = fread((char *)arrptr,sizeof(double),npts*2,fp)) != npts*2) {
         fprintf(stderr,"   ***Binary read error of data array!\n");
         exit(1);
      }

      /* Convert and store the real data points */
      for (i=0; i<npts*2; i=i+2) {
         pt.x = arrptr[i];
         pt.y = arrptr[i+1];
         if (xlog) pt.x = CNconv_to_log(pt.x);
         if (ylog) pt.y = CNconv_to_log(pt.y);
         pt.z = 0.0;
         CNinsert_tailnode(&pt, Cptr);
      }

      /* free the 1D array */
      CNfree_1D_double_array(arrptr);
   }

   /* Store the entire dataset in a plotset */
   Pptr = CNmake_plotset(filename,CN_PLOT2D,
                xmin,xmax,ymin,ymax,zmin,zmax,
                0.0,0,1,0,ID);
   Pptr->curvehead = curvehead;
   Pptr->curvetail = curvetail;

   /* return */
   return(Pptr);
}

