/*
 * round.c - useful routines for rounding
 */

#include <math.h>
#include <stdio.h>
 
#define SMALL      1.0e-8
#define VERY_SMALL 1.0e-30

/* 
 * round down a number to the decimal place given by power
 * e.g round(1.0256,-2) means round to 2nd decimal -> 1.03
 */
static double round_to_pow(x,power)
double x;
int    power;
{
   double inc, xnew;

   inc  = pow(10.0,(double)power);
   if (x > -SMALL)
      xnew = inc * ((int)(x/inc + 0.5 + SMALL));
   else
      xnew = inc * ((int)(x/inc - 0.5 + SMALL));
   return(xnew);
}

/*
 * round down a number to the nearest decimal place
 * i.e.   0.0145 -> 0.01, 0.017 -> 0.020
 */
double
CNround_to_decimal(x)
double x;
{
   static double round_to_pow();
   double dpower, x2;
   int    power;

   /* error-checking */
   if (fabs(x) < SMALL) {
      return(x);
   }

   dpower  = log10(fabs(x));
   power   = (dpower > SMALL) ?  (int)dpower : (int)dpower - 1;
   x2      = round_to_pow(x,power);
   return(x2);
}

/* round down a number */
static double round_down(x,power)
double x;
int    power;
{
   double inc, xnew;
   
   inc  = pow(10.0,(double)power);
   if (x > -VERY_SMALL) 
      xnew = inc * ((int)(x/inc + SMALL));
   else 
      xnew = inc * ((int)(x/inc + SMALL) - 1);
   return(xnew);
}

/* round up a number */
static double round_up(x,power)
double x;
int    power;
{
   double inc, xnew;
   
   inc  = pow(10.0,(double)power);
   if (x > VERY_SMALL) 
      xnew = inc * ((int)(x/inc - SMALL) + 1);
   else  
      xnew = inc * ((int)(x/inc - SMALL));
   return(xnew);
}

/* 
 * autorange on the x-axis 
 */
void CNget_autorange(xmin,xmax,xmin2,xmax2,deltx)
double  xmin,xmax,*xmin2,*xmax2,*deltx;
{
   static double round_down();
   static double round_up();
   double dpower;
   int    power;

   /* error-checking */
   if (fabs(xmax-xmin) < VERY_SMALL) {
      fprintf(stderr,"Log(zero) Error!  xmin = xmax = %g!\n",xmin);
      *xmin2 = xmin;
      *xmax2 = xmax;
   } else {
      /* round down xmin and round up xmax */
      dpower = log10(xmax-xmin+VERY_SMALL);
      power  = (dpower > -SMALL) ? (int)dpower : (int)dpower - 1;
      *xmin2 = round_down(xmin,power);
      *xmax2 = round_up(xmax,power);
      *deltx = pow(10.0,(double)power);
   }
}

/* round down a number automatically to the nearest log scale */
double CNauto_round_down(x)
double x;
{
   double x2, dpower;
   int    power;

   /* error-checking */
   if (fabs(x) < VERY_SMALL) {
      fprintf(stderr,"Log(zero) Error!  x = %g!\n",x);
      exit(-1);
   } 

   /* round down xmin and round up xmax */
   dpower = log10(x+VERY_SMALL);   
   power  = (dpower > -SMALL) ? (int)dpower : (int)dpower - 1;
   x2     = pow(10.0,(double)power);
   return(x2);
}

/* round up a number automatically to the nearest log scale */
double CNauto_round_up(x)
double x;
{
   double x2, dpower;
   int    power;

   /* error-checking */
   if (fabs(x) < VERY_SMALL) {
      fprintf(stderr,"Log(zero) Error!  x = %g!\n",x);
      exit(-1);
   } 

   /* round down xmin and round up xmax */
   dpower = log10(x);   
   power  = (dpower > -SMALL) ? (int)dpower : (int)dpower - 1;
   if ((fabs(dpower - (double)power)) < SMALL)
      x2     = pow(10.0,(double)power);
   else 
      x2     = pow(10.0,(double)(power+1));
   return(x2);
}


/* Take the base10 log of a number */
double CNconv_to_log(x)
double x;
{
   if (x <= 0.0) {
      fprintf(stderr,"Error! Negative values (%f) \
are not acceptable for log plots!\n",x);
      exit(-1);
   }
   return(log10(x));
}

